# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for plugin operations.

This module contains a set of tests that validate the CRUD operations and additional
functionalities for the plugin entity. The tests ensure that the plugins can be
registered, renamed, and deleted as expected through the REST API.
"""
import textwrap
from typing import Any, List

import pytest
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from werkzeug.test import TestResponse

from dioptra.restapi.routes import (
    V1_PLUGIN_FILES_ROUTE,
    V1_PLUGIN_PARAMETER_TYPES_ROUTE,
    V1_PLUGINS_ROUTE,
    V1_ROOT,
)

from ..lib import actions, asserts, helpers

# -- Actions ---------------------------------------------------------------------------


def modify_plugin(
    client: FlaskClient,
    plugin_id: int,
    new_name: str,
    new_description: str,
) -> TestResponse:
    """Rename a plugin using the API.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin to rename.
        new_name: The new name to assign to the plugin.
        new_description: The new description to assign to the plugin.

    Returns:
        The response from the API.
    """
    payload: dict[str, Any] = {"name": new_name, "description": new_description}

    return client.put(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}",
        json=payload,
        follow_redirects=True,
    )


def delete_plugin_with_id(
    client: FlaskClient,
    plugin_id: int,
) -> TestResponse:
    """Delete a plugin using the API.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin to delete.

    Returns:
        The response from the API.
    """
    return client.delete(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}",
        follow_redirects=True,
    )


def modify_plugin_file(
    client: FlaskClient,
    plugin_id: int,
    plugin_file_id: int,
    new_name: str,
    new_contents: str,
    new_description: str,
    new_tasks: list[dict[str, Any]] | None = None,
) -> TestResponse:
    """Rename a plugin file using the API.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin to with plugin files.
        plugin_file_id: The id of the plugin file to rename.
        new_name: The new name to assign to the plugin file.
        new_description: The new description to assign to the plugin file.
        new_tasks: The new tasks to assign to the plugin file.

    Returns:
        The response from the API.
    """
    payload: dict[str, Any] = {
        "filename": new_name,
        "description": new_description,
        "contents": new_contents,
        "tasks": new_tasks or [],
    }

    return client.put(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}/files/{plugin_file_id}",
        json=payload,
        follow_redirects=True,
    )


def delete_plugin_file_with_id(
    client: FlaskClient,
    plugin_id: int,
    plugin_file_id: int,
) -> TestResponse:
    """Delete a plugin file using the API.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin with files.
        plugin_file_id: The id of the plugin file to delete.

    Returns:
        The response from the API.
    """
    return client.delete(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}/files/{plugin_file_id}",
        follow_redirects=True,
    )


def delete_all_plugin_files(client: FlaskClient, plugin_id: int) -> TestResponse:
    """Delete all plugin files for a plugin using the API.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin with files to delete.

    Returns:
        The response from the API.
    """
    return client.delete(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}/files",
        follow_redirects=True,
    )


# -- Assertions Plugins ----------------------------------------------------------------


def assert_plugin_response_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    """Assert that plugin response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """
    expected_keys = {
        "id",
        "snapshot",
        "group",
        "user",
        "createdOn",
        "lastModifiedOn",
        "latestSnapshot",
        "hasDraft",
        "name",
        "description",
        "files",
        "tags",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["id"], int)
    assert isinstance(response["snapshot"], int)
    assert isinstance(response["name"], str)
    assert isinstance(response["description"], str)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)
    assert isinstance(response["latestSnapshot"], bool)
    assert isinstance(response["hasDraft"], bool)

    assert response["name"] == expected_contents["name"]
    assert response["description"] == expected_contents["description"]

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])

    # Validate the UserRef structure
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]

    # Validate the GroupRef structure
    assert isinstance(response["group"]["id"], int)
    assert isinstance(response["group"]["name"], str)
    assert isinstance(response["group"]["url"], str)
    assert response["group"]["id"] == expected_contents["group_id"]

    # Validate the TagRef structure
    for tag in response["tags"]:
        assert isinstance(tag["id"], int)
        assert isinstance(tag["name"], str)
        assert isinstance(tag["url"], str)

    # Validate the PluginFileRef structure
    for file in response["files"]:
        assert isinstance(file["id"], int)
        assert isinstance(file["url"], str)
        assert isinstance(file["plugin_id"], int)
        assert isinstance(file["filename"], str)


def assert_retrieving_plugin_by_id_works(
    client: FlaskClient,
    plugin_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving a plugin by id works.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_plugins_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    group_id: int | None = None,
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all plugins works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        group_id: The group ID used in query parameters.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["group_id"] = group_id

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_sorting_plugin_works(
    client: FlaskClient,
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Assert that plugins can be sorted by column ascending/descending.

    Args:
        client: The Flask test client.
        expected: The expected order of plugins ids after sorting.
            See test_plugin_sort for expected orders.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["sortBy"] = sortBy
    query_string["descending"] = descending

    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )

    response_data = response.get_json()
    plugin_ids = [plugin["id"] for plugin in response_data["data"]]

    assert response.status_code == 200 and plugin_ids == expected


def assert_registering_existing_plugin_name_fails(
    client: FlaskClient, name: str, group_id: int
) -> None:
    """Assert that registering a plugin with an existing name fails.

    Args:
        client: The Flask test client.
        name: The name to assign to the new plugin.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = actions.register_plugin(
        client, name=name, description="", group_id=group_id
    )
    assert response.status_code == 409


def assert_plugin_name_matches_expected_name(
    client: FlaskClient, plugin_id: int, expected_name: str
) -> None:
    """Assert that the name of a plugin matches the expected name.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin to retrieve.
        expected_name: The expected name of the plugin.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            plugin does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["name"] == expected_name


def assert_plugin_is_not_found(
    client: FlaskClient,
    plugin_id: int,
) -> None:
    """Assert that a plugin is not found.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin to retrieve.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


def assert_cannot_rename_plugin_with_existing_name(
    client: FlaskClient,
    plugin_id: int,
    existing_name: str,
    existing_description: str,
) -> None:
    """Assert that renaming a plugin with an existing name fails.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin to rename.
        name: The name of an existing plugin.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = modify_plugin(
        client=client,
        plugin_id=plugin_id,
        new_name=existing_name,
        new_description=existing_description,
    )
    assert response.status_code == 409


# -- Assertions Plugin Files -----------------------------------------------------------


def assert_plugin_file_response_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    """Assert that plugin file response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """
    expected_keys = {
        "id",
        "snapshot",
        "group",
        "user",
        "createdOn",
        "lastModifiedOn",
        "latestSnapshot",
        "hasDraft",
        "filename",
        "description",
        "contents",
        "tasks",
        "tags",
        "plugin",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["id"], int)
    assert isinstance(response["snapshot"], int)
    assert isinstance(response["filename"], str)
    assert isinstance(response["description"], str)
    assert isinstance(response["contents"], str)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)
    assert isinstance(response["latestSnapshot"], bool)
    assert isinstance(response["hasDraft"], bool)

    assert response["filename"] == expected_contents["filename"]
    assert response["description"] == expected_contents["description"]
    assert response["contents"] == expected_contents["contents"]

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])

    # Validate the UserRef structure
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]

    # Validate the GroupRef structure
    assert isinstance(response["group"]["id"], int)
    assert isinstance(response["group"]["name"], str)
    assert isinstance(response["group"]["url"], str)
    assert response["group"]["id"] == expected_contents["group_id"]

    # Validate the PluginTask structure
    for task in response["tasks"]:
        assert isinstance(task["name"], str)

        # Validate PluginTaskParameter Structure for inputs and outputs
        for param in task["inputParams"] + task["outputParams"]:
            assert isinstance(param["name"], str)

            # Validate PluginParameterTypeRef structure
            assert isinstance(param["parameterType"]["id"], int)
            assert isinstance(param["parameterType"]["name"], str)
            assert isinstance(param["parameterType"]["url"], str)

            # Validate the GroupRef structure
            assert isinstance(param["parameterType"]["group"]["id"], int)
            assert isinstance(param["parameterType"]["group"]["name"], str)
            assert isinstance(param["parameterType"]["group"]["url"], str)

    assert response["tasks"] == expected_contents["tasks"]


def assert_retrieving_plugin_file_by_id_works(
    client: FlaskClient,
    plugin_id: int,
    plugin_file_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving a plugin file by id works.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin to that contains the file.
        plugin_file_id: The id of the plugin file to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}/files/{plugin_file_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_plugin_files_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    plugin_id: int,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all plugin files works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        plugin_id: The plugin ID the files are registered too.
        group_id: The group ID used in query parameters.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}/files",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_sorting_plugin_file_works(
    client: FlaskClient,
    sortBy: str,
    descending: bool,
    plugin_id: int,
    expected: list[str],
) -> None:
    """Assert that plugin files can be sorted by column ascending/descending.

    Args:
        client: The Flask test client.
        expected: The expected order of plugins ids after sorting.
            See test_plugin_file_sort for expected orders.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["sortBy"] = sortBy
    query_string["descending"] = descending

    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}/files",
        query_string=query_string,
        follow_redirects=True,
    )

    response_data = response.get_json()
    plugin_ids = [file["id"] for file in response_data["data"]]
    assert response.status_code == 200 and plugin_ids == expected


def assert_registering_existing_plugin_filename_fails(
    client: FlaskClient,
    plugin_id: int,
    existing_filename: str,
    contents: str,
    description: str,
) -> None:
    """Assert that registering a plugin file with an existing name fails.

    Args:
        client: The Flask test client.
        plugin_id: The ID of the plugin with files.
        existing_filename: An existing filename to assign to the new plugin file.
        contents: The string that represents the contents of the file as plain text.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = actions.register_plugin_file(
        client,
        plugin_id=plugin_id,
        filename=existing_filename,
        contents=contents,
        description=description,
    )
    assert response.status_code == 409


def assert_plugin_filename_matches_expected_name(
    client: FlaskClient,
    plugin_id: int,
    plugin_file_id: int,
    expected_name: str,
) -> None:
    """Assert that the name of a plugin file matches the expected name.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin with files.
        plugin_file_id: The id of the plugin file to retrieve.
        expected_name: The expected name of the plugin file.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            plugin does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}/files/{plugin_file_id}",
        follow_redirects=True,
    )
    assert (
        response.status_code == 200 and response.get_json()["filename"] == expected_name
    )


def assert_cannot_rename_plugin_file_with_existing_name(
    client: FlaskClient,
    plugin_id: int,
    plugin_file_id: int,
    existing_name: str,
    existing_contents: str,
    existing_description: str,
) -> None:
    """Assert that renaming a plugin file with an existing name fails.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin with files.
        plugin_file_id: The id of the plugin file to rename.
        existing_name: The name of an existing plugin file.
        existing_description: The  current description of the plugin file.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = modify_plugin_file(
        client=client,
        plugin_id=plugin_id,
        plugin_file_id=plugin_file_id,
        new_name=existing_name,
        new_description=existing_description,
        new_contents=existing_contents,
    )
    assert response.status_code == 409


def assert_plugin_file_is_not_found(
    client: FlaskClient,
    plugin_id: int,
    plugin_file_id: int,
) -> None:
    """Assert that a plugin file is not found.

    Args:
        client: The Flask test client.
        plugin_id: The id of the plugin with files.
        plugin_file_id: The id of the plugin file to retrieve.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGINS_ROUTE}/{plugin_id}/files/{plugin_file_id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


# -- Assertions Plugin Tasks -----------------------------------------------------------


def assert_plugin_task_response_contents_matches_expectations(
    response: List[dict[str, Any]], expected_contents: dict[str, Any]
) -> None:
    """Assert that plugin task response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """
    for task in response:
        expected_keys = {
            "name",
            "input_params",
            "output_params",
        }
        assert set(task.keys()) == expected_keys

        # Validate the non-Ref fields
        assert isinstance(task["name"], str)
        assert task["name"] == expected_contents["name"]

        # Validate PluginTaskParameter Structure for inputs and outputs
        for param in task["input_params"] + task["output_params"]:
            assert isinstance(param["number"], int)
            assert isinstance(param["name"], str)

            # Validate PluginParameterTypeRef structure
            assert isinstance(param["plugin_param_type"]["id"], int)
            assert isinstance(param["plugin_param_type"]["name"], str)

            # Validate the GroupRef structure
            assert isinstance(param["plugin_param_type"]["group"]["id"], int)
            assert isinstance(param["plugin_param_type"]["group"]["name"], str)
            assert isinstance(param["plugin_param_type"]["group"]["url"], str)


# -- Tests Plugins ---------------------------------------------------------------------


def test_create_plugin(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that plugins can be correctly registered and retrieved using the API.

    Given an authenticated user, this test validates the following sequence of actions:

    - The user registers a plugin named "my_plugin".
    - The response is valid matches the expected values given the registration request.
    - The user is able to retrieve information about the plugin using the plugin id.
    """
    name = "my_plugin"
    description = "The first plugin."
    user_id = auth_account["id"]
    group_id = auth_account["groups"][0]["id"]
    plugin_response = actions.register_plugin(
        client, name=name, description=description, group_id=group_id
    )
    plugin_expected = plugin_response.get_json()
    assert_plugin_response_contents_matches_expectations(
        response=plugin_expected,
        expected_contents={
            "name": name,
            "description": description,
            "user_id": user_id,
            "group_id": group_id,
        },
    )
    assert_retrieving_plugin_by_id_works(
        client, plugin_id=plugin_expected["id"], expected=plugin_expected
    )


def test_plugin_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
) -> None:
    """Test that all plugins can be retrieved.

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    - A user registers three plugins, "plugin1", "plugin2", "plugin3".
    - The user is able to retrieve a list of all registered plugins.
    - The returned list of plugins matches the full list of registered plugins.
    """
    plugin_expected_list = list(registered_plugins.values())
    assert_retrieving_plugins_works(client, expected=plugin_expected_list)


@pytest.mark.parametrize(
    "sortBy, descending , expected",
    [
        (None, None, ["plugin1", "plugin2", "plugin3"]),
        ("name", True, ["plugin2", "plugin3", "plugin1"]),
        ("name", False, ["plugin1", "plugin3", "plugin2"]),
        ("createdOn", True, ["plugin3", "plugin2", "plugin1"]),
        ("createdOn", False, ["plugin1", "plugin2", "plugin3"]),
    ],
)
def test_plugin_sort(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Test that plugins can be sorted by column.

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    - A user registers three plugins, "plugin_one", "plugin_two", "plugin_three".
    - The user is able to retrieve a list of all registered plugins sorted by a column
      ascending/descending.
    - The returned list of plugins matches the order in the parametrize lists above.
    """

    expected_ids = [
        registered_plugins[expected_name]["id"] for expected_name in expected
    ]
    assert_sorting_plugin_works(client, sortBy, descending, expected=expected_ids)


def test_plugin_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
) -> None:
    """Test that plugins can be queried with a search term.

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    - The user is able to retrieve a list of all registered plugins with a description
      that contains 'plugin'.
    - The returned list of plugins matches the expected matches from the query.
    """
    plugin_expected_list = list(registered_plugins.values())[:2]
    assert_retrieving_plugins_works(
        client,
        expected=plugin_expected_list,
        search="description:*plugin*",
    )


def test_plugin_group_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
) -> None:
    """Test that plugins can retrieved using a group filter.

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    - The user is able to retrieve a list of all registered plugins that are owned
      by the default group.
    - The returned list of plugins matches the expected list owned by the default group.
    """
    plugin_expected_list = list(registered_plugins.values())
    assert_retrieving_plugins_works(
        client, expected=plugin_expected_list, group_id=auth_account["groups"][0]["id"]
    )


def test_cannot_register_existing_plugin_name(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
) -> None:
    """Test that registering a plugin with an existing name fails.

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    - The user attempts to register a second plugin with the same name.
    - The request fails with an appropriate error message and response code.
    """
    existing_plugin = registered_plugins["plugin1"]
    assert_registering_existing_plugin_name_fails(
        client,
        name=existing_plugin["name"],
        group_id=existing_plugin["group"]["id"],
    )


def test_rename_plugin(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
) -> None:
    """Test that a plugin can be renamed.

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    - The user issues a request to change the name of a plugin.
    - The user retrieves information about the same plugin and it reflects the name
      change.
    - The user issues a request to change the name of a plugin to an existing plugin's
      name.
    - The request fails with an appropriate error message and response code.
    """
    updated_plugin_name = "new_name"
    plugin_to_rename = registered_plugins["plugin1"]
    existing_plugin = registered_plugins["plugin2"]

    modified_plugin = modify_plugin(
        client,
        plugin_id=plugin_to_rename["id"],
        new_name=updated_plugin_name,
        new_description=plugin_to_rename["description"],
    ).get_json()
    assert_plugin_name_matches_expected_name(
        client, plugin_id=plugin_to_rename["id"], expected_name=updated_plugin_name
    )

    plugin_expected_list = [
        modified_plugin,
        registered_plugins["plugin2"],
        registered_plugins["plugin3"],
    ]
    assert_retrieving_plugins_works(client, expected=plugin_expected_list)

    modified_plugin = modify_plugin(
        client,
        plugin_id=plugin_to_rename["id"],
        new_name=updated_plugin_name,
        new_description=plugin_to_rename["description"],
    ).get_json()
    assert_plugin_name_matches_expected_name(
        client, plugin_id=plugin_to_rename["id"], expected_name=updated_plugin_name
    )
    assert_cannot_rename_plugin_with_existing_name(
        client,
        plugin_id=plugin_to_rename["id"],
        existing_name=existing_plugin["name"],
        existing_description=plugin_to_rename["description"],
    )


def test_delete_plugin_by_id(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
) -> None:
    """Test that a plugin can be deleted by referencing its id.

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    - The user deletes a plugin by referencing its id.
    - The user attempts to retrieve information about the deleted plugin.
    - The request fails with an appropriate error message and response code.
    """
    plugin_to_delete = registered_plugins["plugin1"]
    delete_plugin_with_id(client, plugin_id=plugin_to_delete["id"])
    assert_plugin_is_not_found(client, plugin_id=plugin_to_delete["id"])


# -- Tests Plugin Files ----------------------------------------------------------------


def test_register_plugin_file(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that plugin files can be correctly registered to a plugin and retrieved
    using the API.

    Given an authenticated user and a registered plugin, this test validates the
    following sequence of actions:

    - The user registers a plugin file named "my_plugin_file.py".
    - The response is valid matches the expected values given the registration request.
    - The user is able to retrieve information about the plugin using the plugin id.
    """
    registered_plugin = registered_plugins["plugin1"]
    filename = "my_plugin_file.py"
    description = "The first plugin file."
    contents = textwrap.dedent(
        """from dioptra import pyplugs

        @pyplugs.register
        def hello_world(given_name: str, family_name: str) -> str:
            return f"Hello, {given_name} {family_name}!"
        """
    )
    user_id = auth_account["id"]
    string_parameter_type = registered_plugin_parameter_types["plugin_param_type3"]
    tasks = [
        {
            "name": "hello_world",
            "inputParams": [
                {"name": "given_name", "parameterType": string_parameter_type["id"]},
                {"name": "family_name", "parameterType": string_parameter_type["id"]},
            ],
            "outputParams": [
                {
                    "name": "hello_world_message",
                    "parameterType": string_parameter_type["id"],
                }
            ],
        }
    ]
    string_url = (
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{string_parameter_type['id']}"
    )
    expected_tasks = [
        {
            "name": "hello_world",
            "inputParams": [
                {
                    "name": "given_name",
                    "parameterType": {
                        "id": string_parameter_type["id"],
                        "group": string_parameter_type["group"],
                        "url": string_url,
                        "name": string_parameter_type["name"],
                    },
                    "required": True,
                },
                {
                    "name": "family_name",
                    "parameterType": {
                        "id": string_parameter_type["id"],
                        "group": string_parameter_type["group"],
                        "url": string_url,
                        "name": string_parameter_type["name"],
                    },
                    "required": True,
                },
            ],
            "outputParams": [
                {
                    "name": "hello_world_message",
                    "parameterType": {
                        "id": string_parameter_type["id"],
                        "group": string_parameter_type["group"],
                        "url": string_url,
                        "name": string_parameter_type["name"],
                    },
                }
            ],
        }
    ]

    plugin_file_response = actions.register_plugin_file(
        client,
        plugin_id=registered_plugin["id"],
        filename=filename,
        description=description,
        contents=contents,
        tasks=tasks,
    )
    plugin_file_expected = plugin_file_response.get_json()

    assert_plugin_file_response_contents_matches_expectations(
        response=plugin_file_expected,
        expected_contents={
            "filename": filename,
            "description": description,
            "contents": contents,
            "user_id": user_id,
            "group_id": registered_plugin["group"]["id"],
            "tasks": expected_tasks,
        },
    )
    assert_retrieving_plugin_file_by_id_works(
        client,
        plugin_id=registered_plugin["id"],
        plugin_file_id=plugin_file_expected["id"],
        expected=plugin_file_expected,
    )


def test_plugin_file_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
) -> None:
    """Test that all plugin files can be retrieved.

    Given an authenticated user and a registered plugin with plugin files, this test
    validates the following sequence of actions:

    - A user registers three plugin files, "plugin_file1", "plugin_file2",
      "plugin_file3" to the plugin.
    - The user is able to retrieve a list of all registered plugin files.
    - The returned list of plugin files matches the full list of registered plugin
      files.
    """
    registered_plugin = registered_plugin_with_files["plugin"]
    plugin_file1_expected = registered_plugin_with_files["plugin_file1"]
    plugin_file2_expected = registered_plugin_with_files["plugin_file2"]
    plugin_file3_expected = registered_plugin_with_files["plugin_file3"]
    plugin_file_expected_list = [
        plugin_file1_expected,
        plugin_file2_expected,
        plugin_file3_expected,
    ]

    assert_retrieving_plugin_files_works(
        client,
        plugin_id=registered_plugin["id"],
        expected=plugin_file_expected_list,
    )


@pytest.mark.parametrize(
    "sortBy, descending , expected",
    [
        (None, None, ["plugin_file1", "plugin_file2", "plugin_file3"]),
        ("filename", True, ["plugin_file2", "plugin_file3", "plugin_file1"]),
        ("filename", False, ["plugin_file1", "plugin_file3", "plugin_file2"]),
        ("createdOn", True, ["plugin_file3", "plugin_file2", "plugin_file1"]),
        ("createdOn", False, ["plugin_file1", "plugin_file2", "plugin_file3"]),
    ],
)
def test_plugin_file_sort(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Test that plugin files can be sorted by column.

    Given an authenticated user and registered plugin files, this test validates the
    following sequence of actions:

    - A user registers three plugin files:
      "plugin_file_one.py",
      "plugin_file_two.py",
      "plugin_file_three.py".
    - The user is able to retrieve a list of all registered plugins files sorted by a
      column ascending/descending.
    - The returned list of plugin files matches the order in the parametrize lists
      above.
    """

    expected_ids = [
        registered_plugin_with_files[expected_name]["id"] for expected_name in expected
    ]
    assert_sorting_plugin_file_works(
        client,
        sortBy,
        descending,
        plugin_id=registered_plugin_with_files["plugin"]["id"],
        expected=expected_ids,
    )


def test_plugin_file_delete_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
) -> None:
    """Test that all plugin files for a plugin can be deleted.

    Given an authenticated user and a registered plugin with plugin files, this
    test validates the following sequence of actions:

    - The user deletes all plugin files by referencing the plugin id.
    - The user attempts to retrieve information about the deleted plugin files.
    - The returned list of plugin files is empty.
    """
    registered_plugin = registered_plugin_with_files["plugin"]

    plugin_file_expected_list: list[dict[str, Any]] = []

    delete_all_plugin_files(client, plugin_id=registered_plugin["id"])

    assert_retrieving_plugin_files_works(
        client,
        plugin_id=registered_plugin["id"],
        expected=plugin_file_expected_list,
    )


def test_cannot_register_existing_plugin_filename(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
) -> None:
    """Test that registering a plugin file with an existing name fails.

    Given an authenticated user and a registered plugin with plugin files, this
    test validates the following sequence of actions:

    - The user attempts to register a second plugin file with the same name.
    - The request fails with an appropriate error message and response code.
    """
    registered_plugin = registered_plugin_with_files["plugin"]
    existing_plugin_file = registered_plugin_with_files["plugin_file1"]

    assert_registering_existing_plugin_filename_fails(
        client,
        plugin_id=registered_plugin["id"],
        existing_filename=existing_plugin_file["filename"],
        contents=existing_plugin_file["contents"],
        description=existing_plugin_file["description"],
    )


def test_rename_plugin_file(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
) -> None:
    """Test that a plugin file can be renamed.

    Given an authenticated user and a registered plugin with plugin files, this test
    validates the following sequence of actions:

    - The user issues a request to change the name of a plugin file.
    - The user retrieves information about the same plugin file and it reflects the name
      change.
    - The user issues a request to change the name of a plugin file to an existing
      plugin file's name.
    - The request fails with an appropriate error message and response code.
    """
    registered_plugin = registered_plugin_with_files["plugin"]
    plugin_file_to_rename = registered_plugin_with_files["plugin_file1"]
    existing_plugin_file = registered_plugin_with_files["plugin_file2"]
    updated_plugin_filename = "new_name.py"

    modify_plugin_file(
        client,
        plugin_id=registered_plugin["id"],
        plugin_file_id=plugin_file_to_rename["id"],
        new_name=updated_plugin_filename,
        new_contents=plugin_file_to_rename["contents"],
        new_description=plugin_file_to_rename["description"],
    )
    assert_plugin_filename_matches_expected_name(
        client,
        plugin_id=registered_plugin["id"],
        plugin_file_id=plugin_file_to_rename["id"],
        expected_name=updated_plugin_filename,
    )
    assert_cannot_rename_plugin_file_with_existing_name(
        client,
        plugin_id=registered_plugin["id"],
        plugin_file_id=plugin_file_to_rename["id"],
        existing_name=existing_plugin_file["filename"],
        existing_contents=existing_plugin_file["contents"],
        existing_description=plugin_file_to_rename["description"],
    )


def test_delete_plugin_file_by_id(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
) -> None:
    """Test that a plugin file can be deleted by referencing its id.

    Given an authenticated user and a registered plugin with plugin files, this
    test validates the following sequence of actions:

    - The user deletes a plugin file by referencing its id.
    - The user attempts to retrieve information about the deleted plugin file.
    - The request fails with an appropriate error message and response code.
    """
    registered_plugin = registered_plugin_with_files["plugin"]
    plugin_file_to_delete = registered_plugin_with_files["plugin_file1"]

    delete_plugin_file_with_id(
        client,
        plugin_id=registered_plugin["id"],
        plugin_file_id=plugin_file_to_delete["id"],
    )
    assert_plugin_file_is_not_found(
        client,
        plugin_id=registered_plugin["id"],
        plugin_file_id=plugin_file_to_delete["id"],
    )


# -- Tests Plugin Drafts ---------------------------------------------------------------


def test_manage_existing_plugin_draft(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
) -> None:
    """Test that a draft of an existing plugin can be created and managed by the user

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    - The user creates a draft of an existing plugin
    - The user retrieves information about the draft and gets the expected response
    - The user attempts to create another draft of the same existing plugin
    - The request fails with an appropriate error message and response code.
    - The user modifies the name of the plugin in the draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    plugin = registered_plugins["plugin1"]
    name = "draft"
    new_name = "draft2"
    description = "description"

    # test creation
    payload = {"name": name, "description": description}
    expected = {
        "user_id": auth_account["id"],
        "group_id": plugin["group"]["id"],
        "resource_id": plugin["id"],
        "resource_snapshot_id": plugin["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.create_existing_resource_draft(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=plugin["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)
    asserts.assert_retrieving_draft_by_resource_id_works(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=plugin["id"],
        expected=response,
    )
    asserts.assert_creating_another_existing_draft_fails(
        client, resource_route=V1_PLUGINS_ROUTE, resource_id=plugin["id"]
    )

    # test modification
    payload = {"name": new_name, "description": description}
    expected = {
        "user_id": auth_account["id"],
        "group_id": plugin["group"]["id"],
        "resource_id": plugin["id"],
        "resource_snapshot_id": plugin["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.modify_existing_resource_draft(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=plugin["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)

    # test deletion
    actions.delete_existing_resource_draft(
        client, resource_route=V1_PLUGINS_ROUTE, resource_id=plugin["id"]
    )
    asserts.assert_existing_draft_is_not_found(
        client, resource_route=V1_PLUGINS_ROUTE, resource_id=plugin["id"]
    )


def test_manage_new_plugin_drafts(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that drafts of plugin can be created and managed by the user

    Given an authenticated user, this test validates the following sequence of actions:

    - The user creates two plugin drafts
    - The user retrieves information about the drafts and gets the expected response
    - The user modifies the description of the plugin in the first draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the first draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    group_id = auth_account["groups"][0]["id"]
    drafts: dict[str, Any] = {
        "draft1": {"name": "plugin1", "description": "new plugin"},
        "draft2": {"name": "plugin2", "description": None},
    }

    # test creation
    draft1_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft1"],
    }
    draft1_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        group_id=group_id,
        payload=drafts["draft1"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft1_response, draft1_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        draft_id=draft1_response["id"],
        expected=draft1_response,
    )
    draft2_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft2"],
    }
    draft2_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        group_id=group_id,
        payload=drafts["draft2"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft2_response, draft2_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        draft_id=draft2_response["id"],
        expected=draft2_response,
    )
    asserts.assert_retrieving_drafts_works(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        expected=[draft1_response, draft2_response],
    )

    # test modification
    draft1_mod = {"name": "draft1", "description": "new description"}
    draft1_mod_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": draft1_mod,
    }
    response = actions.modify_new_resource_draft(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        draft_id=draft1_response["id"],
        payload=draft1_mod,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        response, draft1_mod_expected
    )

    # test deletion
    actions.delete_new_resource_draft(
        client, resource_route=V1_PLUGINS_ROUTE, draft_id=draft1_response["id"]
    )
    asserts.assert_new_draft_is_not_found(
        client, resource_route=V1_PLUGINS_ROUTE, draft_id=draft1_response["id"]
    )


def test_manage_existing_plugin_file_draft(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
) -> None:
    """Test that a draft of an existing plugin file can be created and managed by the
    user

    Given an authenticated user and registered plugin files, this test validates the
    following sequence of actions:

    - The user creates a draft of an existing plugin file
    - The user retrieves information about the draft and gets the expected response
    - The user attempts to create another draft of the same existing plugin file
    - The request fails with an appropriate error message and response code.
    - The user modifies the name of the plugin file in the draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    plugin_id = registered_plugin_with_files["plugin"]["id"]
    resource_route = f"{V1_PLUGINS_ROUTE}/{plugin_id}/{V1_PLUGIN_FILES_ROUTE}"
    plugin_file = registered_plugin_with_files["plugin_file1"]
    filename = "main.py"
    new_filename = "hello_world.py"
    description = "hello world plugin"
    contents = textwrap.dedent(
        """from dioptra import pyplugs

        @pyplugs.register
        def hello_world(name: str) -> str:
            return f"Hello, {name}!"
        """
    )

    # test creation
    payload = {"filename": filename, "description": description, "contents": contents}
    expected = {
        "user_id": auth_account["id"],
        "group_id": plugin_file["group"]["id"],
        "resource_id": plugin_file["id"],
        "resource_snapshot_id": plugin_file["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.create_existing_resource_draft(
        client,
        resource_route=resource_route,
        resource_id=plugin_file["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)
    asserts.assert_retrieving_draft_by_resource_id_works(
        client,
        resource_route=resource_route,
        resource_id=plugin_file["id"],
        expected=response,
    )
    asserts.assert_creating_another_existing_draft_fails(
        client, resource_route=resource_route, resource_id=plugin_file["id"]
    )

    # test modification
    payload = {
        "filename": new_filename,
        "description": description,
        "contents": contents,
    }
    expected = {
        "user_id": auth_account["id"],
        "group_id": plugin_file["group"]["id"],
        "resource_id": plugin_file["id"],
        "resource_snapshot_id": plugin_file["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.modify_existing_resource_draft(
        client,
        resource_route=resource_route,
        resource_id=plugin_file["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)

    # test deletion
    actions.delete_existing_resource_draft(
        client, resource_route=resource_route, resource_id=plugin_file["id"]
    )
    asserts.assert_existing_draft_is_not_found(
        client, resource_route=resource_route, resource_id=plugin_file["id"]
    )


def test_manage_new_plugin_file_drafts(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
) -> None:
    """Test that drafts of plugin_file can be created and managed by the user

    Given an authenticated user, this test validates the following sequence of actions:

    - The user creates two plugin_file drafts
    - The user retrieves information about the drafts and gets the expected response
    - The user modifies the description of the plugin_file in the first draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the first draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    plugin_id = registered_plugins["plugin1"]["id"]
    resource_route = f"{V1_PLUGINS_ROUTE}/{plugin_id}/{V1_PLUGIN_FILES_ROUTE}"

    group_id = auth_account["groups"][0]["id"]
    contents = textwrap.dedent(
        """from dioptra import pyplugs

        @pyplugs.register
        def hello_world(name: str) -> str:
            return f"Hello, {name}!"
        """
    )
    drafts: dict[str, Any] = {
        "draft1": {
            "filename": "plugin_file1.py",
            "description": "new plugin_file",
            "contents": contents,
        },
        "draft2": {
            "filename": "plugin_file2.py",
            "description": None,
            "contents": contents,
        },
    }

    # test creation
    draft1_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft1"],
    }
    draft1_response = actions.create_new_resource_draft(
        client,
        resource_route=resource_route,
        group_id=None,
        payload=drafts["draft1"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft1_response, draft1_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=resource_route,
        draft_id=draft1_response["id"],
        expected=draft1_response,
    )
    draft2_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft2"],
    }
    draft2_response = actions.create_new_resource_draft(
        client,
        resource_route=resource_route,
        group_id=None,
        payload=drafts["draft2"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft2_response, draft2_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=resource_route,
        draft_id=draft2_response["id"],
        expected=draft2_response,
    )
    asserts.assert_retrieving_drafts_works(
        client,
        resource_route=resource_route,
        expected=[draft1_response, draft2_response],
    )

    # test modification
    draft1_mod = {
        "filename": "draft_plugin.py",
        "description": "new description",
        "contents": contents,
    }
    draft1_mod_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": draft1_mod,
    }
    response = actions.modify_new_resource_draft(
        client,
        resource_route=resource_route,
        draft_id=draft1_response["id"],
        payload=draft1_mod,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        response, draft1_mod_expected
    )

    # test deletion
    actions.delete_new_resource_draft(
        client, resource_route=resource_route, draft_id=draft1_response["id"]
    )
    asserts.assert_new_draft_is_not_found(
        client, resource_route=resource_route, draft_id=draft1_response["id"]
    )


def test_manage_plugin_snapshots(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
) -> None:
    """Test that different snapshots of a plugin can be retrieved by the user.

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    - The user modifies a plugin
    - The user retrieves information about the original snapshot of the plugin and gets
      the expected response
    - The user retrieves information about the new snapshot of the plugin and gets the
      expected response
    - The user retrieves a list of all snapshots of the plugin and gets the expected
      response
    """
    plugin_to_rename = registered_plugins["plugin1"]
    modified_plugin = modify_plugin(
        client,
        plugin_id=plugin_to_rename["id"],
        new_name=plugin_to_rename["name"] + "modified",
        new_description=plugin_to_rename["description"],
    ).get_json()
    modified_plugin.pop("hasDraft")
    modified_plugin.pop("files")
    plugin_to_rename.pop("hasDraft")
    plugin_to_rename.pop("files")
    plugin_to_rename["latestSnapshot"] = False
    plugin_to_rename["lastModifiedOn"] = modified_plugin["lastModifiedOn"]
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=plugin_to_rename["id"],
        snapshot_id=plugin_to_rename["snapshot"],
        expected=plugin_to_rename,
    )
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=modified_plugin["id"],
        snapshot_id=modified_plugin["snapshot"],
        expected=modified_plugin,
    )
    expected_snapshots = [plugin_to_rename, modified_plugin]
    asserts.assert_retrieving_snapshots_works(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=plugin_to_rename["id"],
        expected=expected_snapshots,
    )


def test_manage_plugin_file_snapshots(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
) -> None:
    """Test that different snapshots of a plugin file can be retrieved by the user.

    Given an authenticated user and registered plugin with files, this test validates
    the following sequence of actions:

    - The user modifies a plugin file
    - The user retrieves information about the original snapshot of the plugin file and
      gets the expected response
    - The user retrieves information about the new snapshot of the plugin file and gets
       the expected response
    - The user retrieves a list of all snapshots of the plugin file and gets the
      expected response
    """
    plugin_id = registered_plugin_with_files["plugin"]["id"]
    resource_route = f"{V1_PLUGINS_ROUTE}/{plugin_id}/{V1_PLUGIN_FILES_ROUTE}"
    plugin_file_to_rename = registered_plugin_with_files["plugin_file1"]
    contents = textwrap.dedent(
        """from dioptra import pyplugs

        @pyplugs.register
        def hello_world(name: str) -> str:
            return f"Hello, {name}!"
        """
    )

    modified_plugin_file = modify_plugin_file(
        client,
        plugin_id=plugin_id,
        plugin_file_id=plugin_file_to_rename["id"],
        new_name="modified_" + plugin_file_to_rename["filename"],
        new_contents=contents,
        new_description=plugin_file_to_rename["description"],
    ).get_json()
    modified_plugin_file.pop("hasDraft")
    modified_plugin_file.pop("plugin")
    plugin_file_to_rename.pop("hasDraft")
    plugin_file_to_rename.pop("plugin")
    plugin_file_to_rename["latestSnapshot"] = False
    plugin_file_to_rename["lastModifiedOn"] = modified_plugin_file["lastModifiedOn"]
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=resource_route,
        resource_id=plugin_file_to_rename["id"],
        snapshot_id=plugin_file_to_rename["snapshot"],
        expected=plugin_file_to_rename,
    )
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=resource_route,
        resource_id=modified_plugin_file["id"],
        snapshot_id=modified_plugin_file["snapshot"],
        expected=modified_plugin_file,
    )
    expected_snapshots = [plugin_file_to_rename, modified_plugin_file]
    asserts.assert_retrieving_snapshots_works(
        client,
        resource_route=resource_route,
        resource_id=plugin_file_to_rename["id"],
        expected=expected_snapshots,
    )


def test_tag_plugin(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugins: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that tags can applied to plugins.

    Given an authenticated user and registered plugins, this test validates the
    following sequence of actions:

    """
    plugin = registered_plugins["plugin1"]
    tags = [tag["id"] for tag in registered_tags.values()]

    # test append
    response = actions.append_tags(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=plugin["id"],
        tag_ids=[tags[0], tags[1]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1]]
    )
    response = actions.append_tags(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=plugin["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1], tags[2]]
    )

    # test remove
    actions.remove_tag(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=plugin["id"],
        tag_id=tags[1],
    )
    response = actions.get_tags(
        client, resource_route=V1_PLUGINS_ROUTE, resource_id=plugin["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[2]]
    )

    # test modify
    response = actions.modify_tags(
        client,
        resource_route=V1_PLUGINS_ROUTE,
        resource_id=plugin["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[1], tags[2]]
    )

    # test delete
    response = actions.remove_tags(
        client, resource_route=V1_PLUGINS_ROUTE, resource_id=plugin["id"]
    )
    response = actions.get_tags(
        client, resource_route=V1_PLUGINS_ROUTE, resource_id=plugin["id"]
    )


def test_tag_plugin_file(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that tags can applied to plugin files.

    Given an authenticated user and registered plugin files, this test validates the
    following sequence of actions:

    """
    plugin_id = registered_plugin_with_files["plugin"]["id"]
    plugin_file = registered_plugin_with_files["plugin_file1"]
    tags = [tag["id"] for tag in registered_tags.values()]
    resource_route = f"{V1_PLUGINS_ROUTE}/{plugin_id}/{V1_PLUGIN_FILES_ROUTE}"

    # test append
    response = actions.append_tags(
        client,
        resource_route=resource_route,
        resource_id=plugin_file["id"],
        tag_ids=[tags[0], tags[1]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1]]
    )
    response = actions.append_tags(
        client,
        resource_route=resource_route,
        resource_id=plugin_file["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1], tags[2]]
    )

    # test remove
    actions.remove_tag(
        client,
        resource_route=resource_route,
        resource_id=plugin_file["id"],
        tag_id=tags[1],
    )
    response = actions.get_tags(
        client, resource_route=resource_route, resource_id=plugin_file["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[2]]
    )

    # test modify
    response = actions.modify_tags(
        client,
        resource_route=resource_route,
        resource_id=plugin_file["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[1], tags[2]]
    )

    # test delete
    response = actions.remove_tags(
        client, resource_route=resource_route, resource_id=plugin_file["id"]
    )
    response = actions.get_tags(
        client, resource_route=resource_route, resource_id=plugin_file["id"]
    )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing the auth endpoint objects."""
from __future__ import annotations

from marshmallow import Schema, fields


class LoginSchema(Schema):
    """The request and response fields used when logging into an account."""

    username = fields.String(
        attribute="username",
        metadata=dict(description="The account username."),
        required=True,
    )
    password = fields.String(
        attribute="password",
        metadata=dict(description="The account password."),
        load_only=True,
        required=True,
    )
    status = fields.String(
        attribute="status",
        metadata=dict(description="The status of the login request."),
        dump_only=True,
    )


class LogoutSchema(Schema):
    """The response fields used when logging out of an account."""

    username = fields.String(
        attribute="username",
        metadata=dict(description="The account username."),
        dump_only=True,
        required=True,
    )
    everywhere = fields.Bool(
        attribute="everywhere",
        metadata=dict(description="Logout on all devices."),
        load_default=lambda: False,
    )
    status = fields.String(
        attribute="status",
        metadata=dict(description="The status of the logout request."),
        dump_only=True,
    )


class LogoutQueryParametersSchema(Schema):
    """The query parameters schema for logging out of an account."""

    everywhere = fields.Bool(
        attribute="everywhere",
        metadata=dict(description="Logout on all devices."),
        load_default=lambda: False,
    )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
#
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../../task-plugins"))
sys.path.insert(0, os.path.abspath("../../src"))


# -- Project information -----------------------------------------------------

project = "Dioptra"
release = "1.0.1"
version = ".".join(release.split(".")[:2])

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "recommonmark",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_togglebutton",
    "sphinxcontrib.httpdomain",
    "sphinxcontrib.openapi",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "restructuredtext",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "dev-guide", "getting-started/installation.rst"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

add_module_names = False

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
on_rtd = os.environ.get("READTHEDOCS") == "True"
html_theme = "sphinx_book_theme"
# if on_rtd:
#     html_theme = "default"

# else:
#     html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = ["dioptra.css"]

html_js_files = [
    "jquery.visible.js",
    "jquery.leaveNotice-nist.js",
    "applyLeaveNotice.js",
    "smoothNavScroll.js",
]

html_theme_options = {
    "repository_url": "https://github.com/usnistgov/dioptra",
    "repository_branch": "main",
    "use_repository_button": True,
    "home_page_in_toc": True,
    "path_to_docs": "docs/source",
}

# -- Extension configuration -------------------------------------------------

# -- Options for autodoc extension -------------------------------------------

autoclass_content = "class"
autodoc_mock_imports = [
    "alembic",
    "art",
    "botocore",
    "boto3",
    "entrypoints",
    "flask_injector",
    "flask_migrate",
    "flask_sqlalchemy",
    "injector",
    "mlflow",
    "numpy",
    "pandas",
    "passlib",
    "redis",
    "rq",
    "scipy",
    "sklearn",
    "structlog",
    "tensorflow",
]
autodoc_inherit_docstrings = True
autodoc_member_order = "bysource"
autodoc_typehints = "signature"

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    "alembic": ("https://alembic.sqlalchemy.org/en/latest/", None),
    "art": (
        "https://adversarial-robustness-toolbox.readthedocs.io/en/latest/",
        None,
    ),
    "boto3": ("https://boto3.amazonaws.com/v1/documentation/api/latest/", None),
    "click": ("https://click.palletsprojects.com/en/8.1.x/", None),
    "flask": ("https://flask.palletsprojects.com/en/2.1.x/", None),
    "flask_migrate": ("https://flask-migrate.readthedocs.io/en/latest/", None),
    "flask_restx": ("https://flask-restx.readthedocs.io/en/latest/", None),
    "flask_sqlalchemy": ("https://flask-sqlalchemy.palletsprojects.com/en/2.x/", None),
    "injector": ("https://injector.readthedocs.io/en/latest/", None),
    "marshmallow": ("https://marshmallow.readthedocs.io/en/stable/", None),
    "mlflow": ("https://mlflow.org/docs/latest/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "python": ("https://docs.python.org/3/", None),
    "sqlalchemy": ("https://docs.sqlalchemy.org/en/14/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "sklearn": ("https://scikit-learn.org/1.0/", None),
    "structlog": ("https://www.structlog.org/en/stable/", None),
}

# -- Options for napoleon extension ------------------------------------------

napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_use_ivar = True
napoleon_use_param = True
napoleon_use_rtype = True

# -- Options for panels extension --------------------------------------------

panels_add_bootstrap_css = False



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import uuid
from typing import Any, Dict, Optional, Union

import structlog
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()


class MockRQJob(object):
    def __init__(
        self,
        id: str = "4520511d-678b-4966-953e-af2d0edcea32",
        queue: Optional[str] = None,
        timeout: Optional[str] = None,
        cmd_kwargs: Optional[Dict[str, Any]] = None,
        depends_on: Optional[Union[str, MockRQJob]] = None,
    ) -> None:
        LOGGER.info(
            "Mocking rq.job.Job instance",
            id=id,
            queue=queue,
            timeout=timeout,
            cmd_kwargs=cmd_kwargs,
            depends_on=depends_on,
        )
        self.queue = queue
        self.timeout = timeout
        self.cmd_kwargs = cmd_kwargs
        self._id = id
        self._dependency_ids = None

        if depends_on is not None:
            self._dependency_ids = [
                depends_on.id if isinstance(depends_on, MockRQJob) else depends_on
            ]

    @classmethod
    def fetch(cls, id: str, *args, **kwargs) -> MockRQJob:
        LOGGER.info(
            "Mocking rq.job.Job.fetch() function", id=id, args=args, kwargs=kwargs
        )
        return cls(id=id)

    def get_id(self) -> str:
        LOGGER.info("Mocking rq.job.Job.get_id() function")
        return self._id

    def get_status(self) -> str:
        LOGGER.info("Mocking rq.job.Job.get_status() function")
        return "started"

    @property
    def id(self) -> str:
        LOGGER.info("Mocking rq.job.Job.id getter")
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        LOGGER.info("Mocking rq.job.Job.id setter", value=value)
        self._id = value

    @property
    def dependency(self) -> Optional[MockRQJob]:
        LOGGER.info("Mocking rq.job.Job.dependency getter")
        if not self._dependency_ids:
            return None

        return self.fetch(self._dependency_ids[0])


class MockRQQueue(object):
    def __init__(self, *args, **kwargs):
        LOGGER.info("Mocking rq.Queue instance", args=args, kwargs=kwargs)
        self.name = kwargs.get("name") or args[0]
        self.default_timeout = kwargs.get("default_timeout")

    def enqueue(self, *args, **kwargs) -> MockRQJob:
        LOGGER.info("Mocking rq.Queue.enqueue() function", args=args, kwargs=kwargs)
        job_id = kwargs.get("job_id", str(uuid.uuid4()))
        cmd_kwargs = kwargs.get("kwargs")
        depends_on = kwargs.get("depends_on")
        timeout = kwargs.get("timeout")
        return MockRQJob(
            id=job_id,
            queue=self.name,
            timeout=timeout,
            cmd_kwargs=cmd_kwargs,
            depends_on=depends_on,
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from pathlib import Path
from typing import Callable, Iterable, List, Optional, Sequence, Tuple, cast

import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.object_detection.augmentations import (
    ImgAugObjectDetectionAugmentations,
    ObjectDetectionAugmentations,
    PassthroughObjectDetectionAugmentations,
)
from dioptra.sdk.object_detection.bounding_boxes import (
    TensorflowBoundingBoxesBatchedGrid,
)

from .annotations import (
    AnnotationData,
    NumpyAnnotationEncoding,
    PascalVOCAnnotationData,
)
from .images import TensorflowImageData
from .object_detection_data import ObjectDetectionData

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow import Tensor
    from tensorflow.data import Dataset

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class TensorflowObjectDetectionData(ObjectDetectionData):
    def __init__(
        self,
        annotation_data: AnnotationData,
        bounding_boxes_batched_grid: TensorflowBoundingBoxesBatchedGrid,
        image_data: TensorflowImageData,
        augmentations: ObjectDetectionAugmentations,
        image_dimensions: Tuple[int, int, int],
        grid_shape: Tuple[int, int],
        labels: List[str],
        n_classes: int,
        training_directory: Optional[Path] = None,
        validation_directory: Optional[Path] = None,
        testing_directory: Optional[Path] = None,
        batch_size: Optional[int] = 16,
        shuffle_training_data: bool = True,
        images_dirname: str = "images",
        annotations_dirname: str = "annotations",
        seed: Optional[int] = None,
    ) -> None:
        self._annotation_data = annotation_data
        self._bounding_boxes_batched_grid = bounding_boxes_batched_grid
        self._image_data = image_data
        self._augmentations = augmentations
        self._image_dimensions = image_dimensions
        self._grid_shape = grid_shape
        self._labels = labels
        self._n_classes = n_classes
        self._training_directory = training_directory
        self._validation_directory = validation_directory
        self._testing_directory = testing_directory
        self._batch_size = batch_size
        self._shuffle_training_data = shuffle_training_data
        self._images_dirname = images_dirname
        self._annotations_dirname = annotations_dirname
        self._seed = seed

        self._training_annotations_filepaths: list[str] | None = None
        self._training_images_filepaths: list[str] | None = None
        self._validation_annotations_filepaths: list[str] | None = None
        self._validation_images_filepaths: list[str] | None = None
        self._testing_annotations_filepaths: list[str] | None = None
        self._testing_images_filepaths: list[str] | None = None

    @classmethod
    def create(
        cls,
        image_dimensions: Tuple[int, int, int],
        grid_shape: Tuple[int, int],
        labels: Sequence[str],
        training_directory: Path | str | None = None,
        validation_directory: Path | str | None = None,
        testing_directory: Path | str | None = None,
        annotation_format: str = "pascal_voc",
        augmentations: Optional[str] = None,
        batch_size: Optional[int] = 16,
        shuffle_training_data: bool = True,
        images_dirname: str = "images",
        annotations_dirname: str = "annotations",
        augmentations_seed: Optional[int] = None,
        shuffle_seed: Optional[int] = None,
    ) -> TensorflowObjectDetectionData:
        annotation_data_registry: dict[str, Callable[[], PascalVOCAnnotationData]] = (
            dict(
                pascal_voc=lambda: PascalVOCAnnotationData(
                    labels=labels,
                    encoding=NumpyAnnotationEncoding(
                        boxes_dtype="float32", labels_dtype="int32"
                    ),
                ),
            )
        )
        augmentations_registry: dict[
            str, Callable[[], ImgAugObjectDetectionAugmentations]
        ] = dict(
            imgaug_heavy=(
                lambda: ImgAugObjectDetectionAugmentations.use_heavy_augmenters(
                    image_dimensions=image_dimensions[:2], seed=augmentations_seed
                )
            ),
            imgaug_light=(
                lambda: ImgAugObjectDetectionAugmentations.use_light_augmenters(
                    image_dimensions=image_dimensions[:2], seed=augmentations_seed
                )
            ),
            imgaug_minimal=(
                lambda: ImgAugObjectDetectionAugmentations.use_minimal_augmenters(
                    image_dimensions=image_dimensions[:2], seed=augmentations_seed
                )
            ),
        )

        annotation_data_object = annotation_data_registry[annotation_format]()
        augmentations_object = augmentations_registry.get(
            augmentations or "",
            lambda: PassthroughObjectDetectionAugmentations(),
        )()
        bounding_boxes_batched_grid = TensorflowBoundingBoxesBatchedGrid.on_grid_shape(
            grid_shape=grid_shape
        )
        image_data = TensorflowImageData(
            num_channels=image_dimensions[2], dtype=tf.float32
        )
        n_classes = len(labels)
        training_directory = Path(training_directory) if training_directory else None
        validation_directory = (
            Path(validation_directory) if validation_directory else None
        )
        testing_directory = Path(testing_directory) if testing_directory else None

        return TensorflowObjectDetectionData(
            annotation_data=annotation_data_object,
            bounding_boxes_batched_grid=bounding_boxes_batched_grid,
            image_data=image_data,
            augmentations=augmentations_object,
            image_dimensions=image_dimensions,
            grid_shape=grid_shape,
            labels=[x for x in labels],
            n_classes=n_classes,
            training_directory=training_directory,
            validation_directory=validation_directory,
            testing_directory=testing_directory,
            batch_size=batch_size,
            shuffle_training_data=shuffle_training_data,
            images_dirname=images_dirname,
            annotations_dirname=annotations_dirname,
            seed=shuffle_seed,
        )

    @property
    def augmentations(self) -> ObjectDetectionAugmentations:
        return self._augmentations

    @property
    def labels(self) -> List[str]:
        return self._labels

    @property
    def n_classes(self) -> int:
        return self._n_classes

    @property
    def training_dataset(self) -> Optional[Dataset]:
        if self.training_images_directory is None:
            return None

        dataset = self.create_dataset(
            self.training_images_filepaths, self.training_annotations_filepaths
        )
        dataset = self.shuffle(
            dataset,
            batch_size=self._batch_size,
            seed=self._seed,
            skip=not self._shuffle_training_data,
        )
        dataset = self.map_apply(
            dataset, map_fn=self.load_xy_data_factory(training=True)
        )
        dataset = self.batch(dataset, batch_size=self._batch_size)
        dataset = self.map_apply(dataset, map_fn=self._pack_y_elements)
        dataset = self.prefetch(dataset)

        return dataset

    @property
    def validation_dataset(self) -> Optional[Dataset]:
        if self.validation_images_directory is None:
            return None

        dataset = self.create_dataset(
            self.validation_images_filepaths, self.validation_annotations_filepaths
        )
        dataset = self.map_apply(dataset, map_fn=self.load_xy_data_factory())
        dataset = self.batch(dataset, batch_size=self._batch_size)
        dataset = self.map_apply(dataset, map_fn=self._pack_y_elements)
        dataset = self.prefetch(dataset)

        return dataset

    @property
    def testing_dataset(self) -> Optional[Dataset]:
        if self.testing_images_directory is None:
            return None

        dataset = self.create_dataset(
            self.testing_images_filepaths, self.testing_annotations_filepaths
        )
        dataset = self.map_apply(dataset, map_fn=self.load_xy_data_factory())
        dataset = self.batch(dataset, batch_size=self._batch_size)
        dataset = self.map_apply(dataset, map_fn=self._pack_y_elements)
        dataset = self.prefetch(dataset)

        return dataset

    @property
    def training_annotations_directory(self) -> Optional[Path]:
        if self._training_directory is None:
            return None

        return (self._training_directory / self._annotations_dirname).resolve()

    @property
    def training_annotations_filepaths(self) -> Optional[List[str]]:
        if self.training_annotations_directory is None:
            return None

        if self._training_annotations_filepaths is None:
            self._training_annotations_filepaths = self.to_annotations_filepaths(
                annotations_directory=self.training_annotations_directory,
                images_filepaths=self.training_images_filepaths,
            )

        return self._training_annotations_filepaths

    @property
    def training_images_directory(self) -> Optional[Path]:
        if self._training_directory is None:
            return None

        return (self._training_directory / self._images_dirname).resolve()

    @property
    def training_images_filepaths(self) -> Optional[List[str]]:
        if self.training_images_directory is None:
            return None

        if self._training_images_filepaths is None:
            self._training_images_filepaths = sorted(
                [str(x) for x in self.training_images_directory.glob("*")]
            )

        return self._training_images_filepaths

    @property
    def testing_annotations_directory(self) -> Optional[Path]:
        if self._testing_directory is None:
            return None

        return (self._testing_directory / self._annotations_dirname).resolve()

    @property
    def testing_annotations_filepaths(self) -> Optional[List[str]]:
        if self.testing_annotations_directory is None:
            return None

        if self._testing_annotations_filepaths is None:
            self._testing_annotations_filepaths = self.to_annotations_filepaths(
                annotations_directory=self.testing_annotations_directory,
                images_filepaths=self.testing_images_filepaths,
            )

        return self._testing_annotations_filepaths

    @property
    def testing_images_directory(self) -> Optional[Path]:
        if self._testing_directory is None:
            return None

        return (self._testing_directory / self._images_dirname).resolve()

    @property
    def testing_images_filepaths(self) -> Optional[List[str]]:
        if self.testing_images_directory is None:
            return None

        if self._testing_images_filepaths is None:
            self._testing_images_filepaths = sorted(
                [str(x) for x in self.testing_images_directory.glob("*")]
            )

        return self._testing_images_filepaths

    @property
    def validation_annotations_directory(self) -> Optional[Path]:
        if self._validation_directory is None:
            return None

        return (self._validation_directory / self._annotations_dirname).resolve()

    @property
    def validation_annotations_filepaths(self) -> Optional[List[str]]:
        if self.validation_annotations_directory is None:
            return None

        if self._validation_annotations_filepaths is None:
            self._validation_annotations_filepaths = self.to_annotations_filepaths(
                annotations_directory=self.validation_annotations_directory,
                images_filepaths=self.validation_images_filepaths,
            )

        return self._validation_annotations_filepaths

    @property
    def validation_images_directory(self) -> Optional[Path]:
        if self._validation_directory is None:
            return None

        return (self._validation_directory / self._images_dirname).resolve()

    @property
    def validation_images_filepaths(self) -> Optional[List[str]]:
        if self.validation_images_directory is None:
            return None

        if self._validation_images_filepaths is None:
            self._validation_images_filepaths = sorted(
                [str(x) for x in self.validation_images_directory.glob("*")]
            )

        return self._validation_images_filepaths

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def augment_data(
        self, image: Tensor, bboxes: Tensor, labels: Tensor
    ) -> tuple[Tensor, Tensor, Tensor]:
        return cast(
            tuple[Tensor, Tensor, Tensor],
            tf.numpy_function(
                self.augmentations.augment,
                [image, bboxes, labels],
                [tf.float32, tf.float32, tf.int32],
            ),
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def embed_bounding_boxes(self, bboxes_corner: Tensor, bboxes_labels: Tensor):
        n_classes = tf.cast(self.n_classes, tf.int32)

        return self._bounding_boxes_batched_grid.embed(
            bboxes_corner=bboxes_corner,
            bboxes_labels=bboxes_labels,
            n_classes=n_classes,
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.string),
        ]
    )
    def load_image(self, filepath: Tensor) -> Tensor:
        image = self._image_data.read_file(filepath=filepath)
        image = self._image_data.resize(images=image, size=self._image_dimensions[:2])

        return image

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.string),
        ]
    )
    def load_annotations(self, y: Tensor) -> tuple[Tensor, Tensor]:
        return cast(
            tuple[Tensor, Tensor],
            tf.numpy_function(self._annotation_data.get, [y], [tf.float32, tf.int32]),
        )

    def load_xy_data_factory(
        self, training: bool = False
    ) -> Callable[[Tensor, Tensor], tuple[Tensor, Tensor, Tensor, Tensor, Tensor]]:
        @tf.function(
            input_signature=[
                tf.TensorSpec(None, tf.string),
                tf.TensorSpec(None, tf.string),
            ]
        )
        def load_xy_data(
            x: Tensor, y: Tensor
        ) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
            return cast(
                tuple[Tensor, Tensor, Tensor, Tensor, Tensor],
                tf.py_function(
                    self._load_xy_data,
                    [x, y, training],
                    [tf.float32, tf.float32, tf.float32, tf.float32, tf.float32],
                ),
            )

        return cast(
            Callable[[Tensor, Tensor], tuple[Tensor, Tensor, Tensor, Tensor, Tensor]],
            load_xy_data,
        )

    def _load_xy_data(
        self, x: Tensor, y: Tensor, training: bool = False
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
        image = self.load_image(x)
        bboxes, labels = self.load_annotations(y)

        if training:
            image, bboxes, labels = self.augment_data(
                image=image, bboxes=bboxes, labels=labels
            )

        (
            bboxes_cell_xywh_grid,
            bboxes_labels_grid,
            bboxes_object_mask,
            bboxes_no_object_mask,
        ) = self.embed_bounding_boxes(bboxes_corner=bboxes, bboxes_labels=labels)

        return (
            image,
            bboxes_cell_xywh_grid,
            bboxes_labels_grid,
            bboxes_object_mask,
            bboxes_no_object_mask,
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _pack_y_elements(
        self,
        image: Tensor,
        bboxes_cell_xywh_grid: Tensor,
        bboxes_labels_grid: Tensor,
        bboxes_object_mask: Tensor,
        bboxes_no_object_mask: Tensor,
    ) -> Tuple[Tensor, Tuple[Tensor, Tensor, Tensor, Tensor]]:
        return (
            image,
            (
                bboxes_cell_xywh_grid,
                bboxes_labels_grid,
                bboxes_object_mask,
                bboxes_no_object_mask,
            ),
        )

    @staticmethod
    def batch(dataset: Dataset, batch_size: Optional[int] = None) -> Dataset:
        if batch_size is None:
            return dataset

        return dataset.batch(batch_size)

    @staticmethod
    def create_dataset(images_filepaths, annotations_filepaths) -> Dataset:
        return Dataset.from_tensor_slices(
            (
                tf.convert_to_tensor(images_filepaths, dtype=tf.string),
                tf.convert_to_tensor(annotations_filepaths, dtype=tf.string),
            )
        )

    @staticmethod
    def map_apply(
        dataset: Dataset,
        map_fn: Callable[
            [Tensor, Tensor], Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]
        ],
    ) -> Dataset:
        return dataset.map(map_fn)

    @staticmethod
    def prefetch(dataset: Dataset) -> Dataset:
        autotune = tf.data.AUTOTUNE
        return dataset.prefetch(autotune)

    @staticmethod
    def shuffle(
        dataset: Dataset,
        batch_size: Optional[int] = None,
        seed: Optional[int] = None,
        reshuffle_each_iteration: bool = True,
        skip: bool = False,
    ) -> Dataset:
        if skip:
            return dataset

        buffer_size = 8 if batch_size is None else 8 * batch_size

        return dataset.shuffle(
            buffer_size=buffer_size,
            seed=seed,
            reshuffle_each_iteration=reshuffle_each_iteration,
        )

    @staticmethod
    def to_annotations_filepaths(
        annotations_directory: Path, images_filepaths: Iterable[str] | None
    ) -> List[str]:
        annotations_filepaths: List[str] = []

        if images_filepaths is None:
            return annotations_filepaths

        for filepath in images_filepaths:
            annotation_filename = Path(filepath).with_suffix(".xml").name
            annotation_filepath = annotations_directory / annotation_filename
            annotations_filepaths.append(str(annotation_filepath))

        return annotations_filepaths



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from collections.abc import Iterable
from typing import Any, Optional

import dioptra.task_engine.types

from .base import BaseTaskEngineError


class StepError(BaseTaskEngineError):
    """
    An error which can occur or exist within the context of a particular step
    of a task graph.  This class has support for storing the contextual step
    name and producing a better error message.
    """

    def __init__(self, message: str, context_step_name: Optional[str] = None) -> None:
        """
        Initialize this error instance.

        Args:
            message: An error message
            context_step_name: The name of the step which was the context of
                the error, or None.  If None, the step name can be populated
                later, e.g. filled in at a higher stack frame where the info is
                known.
        """

        super().__init__(message)

        # Step name which is the context for this error.
        self.context_step_name = context_step_name

        self.__message = message

    def __str__(self) -> str:
        """
        Override string representation such that it depends on what context we
        have for the error.  With more context, we can produce a better error
        message.

        Returns:
            A composed error message string
        """
        msg_parts = []
        if self.context_step_name:
            msg_parts.append('In step "{}": '.format(self.context_step_name))

        msg_parts.append(self.__message)

        return "".join(msg_parts)


class StepNotFoundError(StepError):
    """A reference to a nonexistent step."""

    def __init__(self, step_name: str, context_step_name: Optional[str] = None) -> None:
        super().__init__("Step not found: " + step_name, context_step_name)

        self.step_name = step_name


class OutputNotFoundError(StepError):
    """A reference to a nonexistent output of an existing step."""

    def __init__(
        self, step_name: str, output_name: str, context_step_name: Optional[str] = None
    ) -> None:
        super().__init__(
            'Unrecognized output of step "{}": {}'.format(step_name, output_name),
            context_step_name,
        )

        self.step_name = step_name
        self.output_name = output_name


class IllegalOutputReferenceError(StepError):
    """A reference to a multi-output step did not name the desired output."""

    def __init__(self, step_name: str, context_step_name: Optional[str] = None) -> None:
        super().__init__(
            "An output name is required when referring to a step with more"
            " than one output: " + step_name,
            context_step_name,
        )

        self.step_name = step_name


class NonIterableTaskOutputError(StepError):
    """
    Task output was defined using a list, but the task invocation did not
    return an iterable value.
    """

    def __init__(self, value: Any, context_step_name: Optional[str] = None) -> None:
        super().__init__(
            "Task output was defined using a list, but the task invocation did"
            " not return an iterable value: {} ({})".format(value, type(value)),
            context_step_name,
        )

        self.value = value


class UnresolvableReferenceError(StepError):
    """
    A reference did not resolve and does not match any global parameter or
    step (so we don't know what the intent was), or the reference was to the
    output of a step which was not declared to produce any output.
    """

    def __init__(
        self, reference_name: str, context_step_name: Optional[str] = None
    ) -> None:
        super().__init__("Unresolvable reference: " + reference_name, context_step_name)

        self.reference_name = reference_name


class TaskPluginNotFoundError(StepError):
    """A reference to a nonexistent task plugin."""

    def __init__(
        self, task_plugin_short_name: str, context_step_name: Optional[str] = None
    ) -> None:
        super().__init__(
            "Task plugin not found: " + task_plugin_short_name, context_step_name
        )

        self.task_plugin_short_name = task_plugin_short_name


class MissingTaskPluginNameError(StepError):
    """
    A step description was malformed: it was missing a task plugin short name.
    """

    def __init__(self, context_step_name: Optional[str] = None) -> None:
        super().__init__("Step is missing a task plugin name", context_step_name)


class MissingGlobalParametersError(BaseTaskEngineError):
    """
    A value could not be obtained for some task graph global parameter(s).
    """

    def __init__(self, parameter_names: Iterable[str]) -> None:
        super().__init__(
            "Missing values for global parameters: " + ", ".join(parameter_names)
        )

        self.parameter_names = parameter_names


class IllegalPluginNameError(BaseTaskEngineError):
    """A task was defined with an illegal plugin name."""

    def __init__(self, plugin_name: str) -> None:
        super().__init__(
            "A task plugin must be a dot-delimited string of at least two"
            " components: " + plugin_name
        )

        self.plugin_name = plugin_name


class StepReferenceCycleError(BaseTaskEngineError):
    """A cycle was detected in the step graph."""

    def __init__(self, cycle: Iterable[str]) -> None:
        super().__init__("Step cycle detected: {}".format(" > ".join(cycle)))

        self.cycle = cycle


class DioptraTypeError(BaseTaskEngineError):
    # Can't name this "TypeError", since that's a builtin Python exception
    # type!
    """
    An error which can occur or exist within the context of a particular
    top-level type definition.  This class has support for storing the
    contextual type name and producing a better error message.
    """

    def __init__(self, message: str, context_type_name: Optional[str] = None) -> None:
        """
        Initialize this error instance.

        :param message: An error message
        :param context_type_name: The name of the type which was the context of
            the error, or None.  If None, the type name can be populated later,
            e.g. filled in at a higher stack frame where the info is known.
        """

        super().__init__(message)

        # Step name which is the context for this error.
        self.context_type_name = context_type_name

        self.__message = message

    def __str__(self) -> str:
        """
        Override string representation such that it depends on what context we
        have for the error.  With more context, we can produce a better error
        message.

        :return: A composed error message string
        """
        msg_parts = []
        if self.context_type_name:
            msg_parts.append(
                'In definition of type "{}": '.format(self.context_type_name)
            )

        msg_parts.append(self.__message)

        return "".join(msg_parts)


class TypeNotFoundError(DioptraTypeError):
    """
    A reference to an unknown type was found.
    """

    def __init__(self, type_name: str, context_type_name: Optional[str] = None) -> None:
        message = "Type not found: " + type_name
        super().__init__(message, context_type_name)

        self.type_name = type_name


class NonSimpleSuperTypeError(DioptraTypeError):
    """
    A simple type definition referenced a super-type which was not simple.
    """

    def __init__(
        self, super_type_name: str, context_type_name: Optional[str] = None
    ) -> None:
        message = "Super-types must be simple: " + super_type_name
        super().__init__(message, context_type_name)

        self.super_type_name = super_type_name


class InvalidKeyTypeError(DioptraTypeError):
    """
    A mapping structured type was defined with an invalid key type.
    """

    def __init__(
        self,
        invalid_key_type: "dioptra.task_engine.types.Type",
        context_type_name: Optional[str] = None,
    ) -> None:
        message = "Invalid key type; require string or integer: {!s}".format(
            invalid_key_type
        )

        super().__init__(message, context_type_name)

        self.invalid_key_type = invalid_key_type


class TooManyTypeStructuresError(DioptraTypeError):
    """
    A type definition included more than one structure type.
    """

    def __init__(
        self,
        structure_types: Iterable["dioptra.task_engine.types.StructureType"],
        context_type_name: Optional[str] = None,
    ) -> None:
        message = (
            "A type definition may include at most one structure type." "  Found: {}"
        ).format(", ".join(structure_type.name for structure_type in structure_types))

        super().__init__(message, context_type_name)

        self.structure_types = structure_types


class InvalidTypeStructureDefinitionError(DioptraTypeError):
    """
    The definition of a structured type was invalid.
    """

    def __init__(self, message: str, context_type_name: Optional[str] = None) -> None:
        super().__init__(message, context_type_name)


class BuiltinTypeRedefinitionError(DioptraTypeError):
    """
    A type definition tried to redefine a builtin type.
    """

    def __init__(self, builtin_type_name: str) -> None:
        message = "Redefinition of builtin types not allowed: {}".format(
            builtin_type_name
        )

        super().__init__(message)

        self.builtin_type_name = builtin_type_name


class AnonymousSimpleTypeError(DioptraTypeError):
    """
    An attempt to create an anonymous type from a null type definition (which
    signifies a simple type).
    """

    def __init__(self) -> None:
        message = (
            "Unable to create an anonymous type from a null type"
            " definition: simple types can't be anonymous."
        )

        super().__init__(message)


class TypeReferenceCycleError(BaseTaskEngineError):
    """A cycle was detected in a set of type definitions."""

    def __init__(self, cycle: Iterable[str]) -> None:
        super().__init__("Type reference cycle detected: {}".format(" > ".join(cycle)))

        self.cycle = cycle



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for model operations.

This module contains a set of tests that validate the CRUD operations and additional
functionalities for the model entity. The tests ensure that the models can be
registered, renamed, deleted, and locked/unlocked as expected through the REST API.
"""

from typing import Any

import pytest
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from werkzeug.test import TestResponse

from dioptra.restapi.routes import V1_MODELS_ROUTE, V1_ROOT

from ..lib import actions, asserts, helpers

# -- Actions ---------------------------------------------------------------------------


def modify_model(
    client: FlaskClient,
    model_id: int,
    new_name: str,
    new_description: str | None,
) -> TestResponse:
    """Rename a model using the API.

    Args:
        client: The Flask test client.
        model_id: The id of the model to rename.
        new_name: The new name to assign to the model.
        new_description: The new description to assign to the model.
        new_artifact_id: The new artifact to assign to the model.

    Returns:
        The response from the API.
    """
    payload = {
        "name": new_name,
        "description": new_description,
    }

    return client.put(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}",
        json=payload,
        follow_redirects=True,
    )


def delete_model_with_id(
    client: FlaskClient,
    model_id: int,
) -> TestResponse:
    """Delete a model using the API.

    Args:
        client: The Flask test client.
        model_id: The id of the model to delete.

    Returns:
        The response from the API.
    """

    return client.delete(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}",
        follow_redirects=True,
    )


def modify_model_version(
    client: FlaskClient,
    model_id: int,
    version_number: int,
    new_description: str | None,
) -> TestResponse:
    """Change the description of a model version using the API.

    Args:
        client: The Flask test client.
        model_id: The id of the model to rename.
        new_description: The new description to assign to the model.

    Returns:
        The response from the API.
    """
    payload = {"description": new_description}

    return client.put(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}/versions/{version_number}",
        json=payload,
        follow_redirects=True,
    )


# -- Assertions ------------------------------------------------------------------------


def assert_model_response_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    """Assert that model response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """
    # Check expected keys
    expected_keys = {
        "id",
        "snapshot",
        "group",
        "user",
        "versions",
        "createdOn",
        "lastModifiedOn",
        "latestSnapshot",
        "hasDraft",
        "name",
        "description",
        "latestVersion",
        "tags",
    }
    assert set(response.keys()) == expected_keys

    # Check basic response types for base resources
    asserts.assert_base_resource_contents_match_expectations(response)

    # Check basic response types for model
    assert isinstance(response["name"], str)
    assert isinstance(response["description"], str)
    assert isinstance(response["versions"], list)

    assert response["name"] == expected_contents["name"]
    assert response["description"] == expected_contents["description"]
    assert response["versions"] == expected_contents["versions"]
    assert response["latestVersion"] == expected_contents["latest_version"]

    # Check Refs for base resources
    asserts.assert_user_ref_contents_matches_expectations(
        user=response["user"], expected_user_id=expected_contents["user_id"]
    )
    asserts.assert_group_ref_contents_matches_expectations(
        group=response["group"], expected_group_id=expected_contents["group_id"]
    )
    asserts.assert_tag_ref_contents_matches_expectations(tags=response["tags"])


def assert_model_version_response_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    """Assert that model version response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """
    expected_keys = {
        "model",
        "artifact",
        "versionNumber",
        "description",
        "createdOn",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["description"], str)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["versionNumber"], int)

    # assert response["versionNumber"] == expected_contents["version_number"]
    assert response["description"] == expected_contents["description"]
    assert response["model"]["id"] == expected_contents["model_id"]
    assert response["artifact"]["id"] == expected_contents["artifact_id"]

    assert helpers.is_iso_format(response["createdOn"])


def assert_retrieving_model_by_id_works(
    client: FlaskClient,
    model_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving a model by id works.

    Args:
        client: The Flask test client.
        model_id: The id of the model to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_models_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    group_id: int | None = None,
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all models works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        group_id: The group ID used in query parameters.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    if group_id is not None:
        query_string["groupId"] = group_id

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_sorting_model_works(
    client: FlaskClient,
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Assert that models can be sorted by column ascending/descending.

    Args:
        client: The Flask test client.
        expected: The expected order of model ids after sorting.
            See test_models_sort for expected orders.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["sortBy"] = sortBy
    query_string["descending"] = descending

    response = client.get(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )

    response_data = response.get_json()
    model_ids = [model["id"] for model in response_data["data"]]

    assert response.status_code == 200 and model_ids == expected


def assert_registering_existing_model_name_fails(
    client: FlaskClient, name: str, group_id: int
) -> None:
    """Assert that registering a model with an existing name fails.

    Args:
        client: The Flask test client.
        name: The name to assign to the new model.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = actions.register_model(
        client, name=name, description="", group_id=group_id
    )
    assert response.status_code == 409


def assert_model_name_matches_expected_name(
    client: FlaskClient, model_id: int, expected_name: str
) -> None:
    """Assert that the name of a model matches the expected name.

    Args:
        client: The Flask test client.
        model_id: The id of the model to retrieve.
        expected_name: The expected name of the model.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            model does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["name"] == expected_name


def assert_model_is_not_found(
    client: FlaskClient,
    model_id: int,
) -> None:
    """Assert that a model is not found.

    Args:
        client: The Flask test client.
        model_id: The id of the model to retrieve.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


def assert_cannot_rename_model_with_existing_name(
    client: FlaskClient,
    model_id: int,
    existing_name: str,
    existing_description: str,
) -> None:
    """Assert that renaming a model with an existing name fails.

    Args:
        client: The Flask test client.
        model_id: The id of the model to rename.
        name: The name of an existing model.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = modify_model(
        client=client,
        model_id=model_id,
        new_name=existing_name,
        new_description=existing_description,
    )
    assert response.status_code == 409


def assert_retrieving_model_version_by_version_number_works(
    client: FlaskClient, model_id: int, version_number: int, expected: dict[str, Any]
) -> None:
    """Assert that retrieving a model version by id works.

    Args:
        client: The Flask test client.
        model_id: The id of the model.
        model_version_id: The id of the model version to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}/versions/{version_number}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_model_versions_works(
    client: FlaskClient,
    model_id: int,
    expected: dict[str, Any],
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """"""
    query_string: dict[str, Any] = {}

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}/versions",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_model_version_description_matches_expected_description(
    client: FlaskClient,
    model_id: int,
    version_number: int,
    expected_description: str,
) -> None:
    """Assert that the description of a model version matches the expected description.

    Args:
        client: The Flask test client.
        model_id: The id of the model.
        version_number: The version number of the model version to retrieve.
        expected_description: The expected description of the model version.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            model does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_MODELS_ROUTE}/{model_id}/versions/{version_number}",
        follow_redirects=True,
    )
    assert (
        response.status_code == 200
        and response.get_json()["description"] == expected_description
    )


# -- Model Tests ---------------------------------------------------------------


def test_create_model(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """"""
    name = "my_model"
    description = "The first model."
    user_id = auth_account["id"]
    group_id = auth_account["groups"][0]["id"]
    model_response = actions.register_model(
        client, name=name, group_id=group_id, description=description
    )

    model_expected = model_response.get_json()
    assert_model_response_contents_matches_expectations(
        response=model_expected,
        expected_contents={
            "name": name,
            "description": description,
            "user_id": user_id,
            "group_id": group_id,
            "versions": [],
            "latest_version": None,
        },
    )

    assert_retrieving_model_by_id_works(
        client, model_id=model_expected["id"], expected=model_expected
    )


def test_model_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
) -> None:
    """Test that all models can be retrieved.

    Given an authenticated user and registered models, this test validates the following
    sequence of actions:

    - A user registers three models, "tensorflow_cpu", "tensorflow_gpu", "pytorch_cpu".
    - The user is able to retrieve a list of all registered models.
    - The returned list of models matches the full list of registered models.
    """
    model_expected_list = list(registered_models.values())
    assert_retrieving_models_works(client, expected=model_expected_list)


@pytest.mark.parametrize(
    "sortBy, descending , expected",
    [
        (None, None, ["model1", "model2", "model3"]),
        ("name", True, ["model1", "model3", "model2"]),
        ("name", False, ["model2", "model3", "model1"]),
        ("createdOn", True, ["model3", "model2", "model1"]),
        ("createdOn", False, ["model1", "model2", "model3"]),
    ],
)
def test_model_sort(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Test that models can be sorted by column.

    Given an authenticated user and registered models, this test validates the following
    sequence of actions:

    - A user registers three models, "my_tensorflow_model", "model2", "model3".
    - The user is able to retrieve a list of all registered models sorted by a column
      ascending/descending.
    - The returned list of models matches the order in the parametrize lists above.
    """

    expected_ids = [
        registered_models[expected_name]["id"] for expected_name in expected
    ]
    assert_sorting_model_works(client, sortBy, descending, expected=expected_ids)


def test_model_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
) -> None:
    """Test that models can be queried with a search term.

    Given an authenticated user and registered models, this test validates the following
    sequence of actions:

    - The user is able to retrieve a list of all registered models with various queries.
    - The returned list of models matches the expected matches from the query.
    """
    model_expected_list = list(registered_models.values())[:2]
    assert_retrieving_models_works(
        client, expected=model_expected_list, search="description:*model*"
    )
    model_expected_list = list(registered_models.values())[:1]
    assert_retrieving_models_works(
        client, expected=model_expected_list, search="*model*, name:*tensorflow*"
    )
    model_expected_list = list(registered_models.values())
    assert_retrieving_models_works(client, expected=model_expected_list, search="*")


def test_model_group_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
) -> None:
    """Test that models can retrieved using a group filter.

    Given an authenticated user and registered models, this test validates the following
    sequence of actions:

    - The user is able to retrieve a list of all registered models that are owned by the
      default group.
    - The returned list of models matches the expected list owned by the default group.
    """
    model_expected_list = list(registered_models.values())
    assert_retrieving_models_works(
        client,
        expected=model_expected_list,
        group_id=auth_account["groups"][0]["id"],
    )


def test_cannot_register_existing_model_name(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
) -> None:
    """Test that registering a model with an existing name fails.

    Given an authenticated user and registered models, this test validates the following
    sequence of actions:

    - The user attempts to register a second model with the same name.
    - The request fails with an appropriate error message and response code.
    """
    existing_model = registered_models["model1"]

    assert_registering_existing_model_name_fails(
        client,
        name=existing_model["name"],
        group_id=existing_model["group"]["id"],
    )


def test_rename_model(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
) -> None:
    """Test that a model can be renamed.

    Given an authenticated user and registered models, this test validates the following
    sequence of actions:

    - The user issues a request to change the name of a model.
    - The user retrieves information about the same model and it reflects the name
      change.
    - The user issues a request to change the name of the model to the existing name.
    - The user retrieves information about the same model and verifies the name remains
      unchanged.
    - The user issues a request to change the name of a model to an existing model's
      name.
    - The request fails with an appropriate error message and response code.
    """
    updated_model_name = "new_model_name"
    model_to_rename = registered_models["model1"]
    existing_model = registered_models["model2"]

    modified_model = modify_model(
        client,
        model_id=model_to_rename["id"],
        new_name=updated_model_name,
        new_description=model_to_rename["description"],
    ).get_json()
    assert_model_name_matches_expected_name(
        client, model_id=model_to_rename["id"], expected_name=updated_model_name
    )
    model_expected_list = [
        modified_model,
        registered_models["model2"],
        registered_models["model3"],
    ]
    assert_retrieving_models_works(client, expected=model_expected_list)

    modified_model = modify_model(
        client,
        model_id=model_to_rename["id"],
        new_name=updated_model_name,
        new_description=model_to_rename["description"],
    ).get_json()
    assert_model_name_matches_expected_name(
        client, model_id=model_to_rename["id"], expected_name=updated_model_name
    )

    assert_cannot_rename_model_with_existing_name(
        client,
        model_id=model_to_rename["id"],
        existing_name=existing_model["name"],
        existing_description=model_to_rename["description"],
    )


def test_delete_model_by_id(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
) -> None:
    """Test that a model can be deleted by referencing its id.

    Given an authenticated user and registered models, this test validates the following
    sequence of actions:

    - The user deletes a model by referencing its id.
    - The user attempts to retrieve information about the deleted model.
    - The request fails with an appropriate error message and response code.
    """
    model_to_delete = registered_models["model1"]

    delete_model_with_id(client, model_id=model_to_delete["id"])
    assert_model_is_not_found(client, model_id=model_to_delete["id"])


def test_manage_existing_model_draft(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
) -> None:
    """Test that a draft of an existing model can be created and managed by the user

    Given an authenticated user and registered models, this test validates the following
    sequence of actions:

    - The user creates a draft of an existing model
    - The user retrieves information about the draft and gets the expected response
    - The user attempts to create another draft of the same existing model
    - The request fails with an appropriate error message and response code.
    - The user modifies the name of the model in the draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    model = registered_models["model1"]
    name = "draft"
    new_name = "draft2"
    description = "description"

    # test creation
    payload = {"name": name, "description": description}
    expected = {
        "user_id": auth_account["id"],
        "group_id": model["group"]["id"],
        "resource_id": model["id"],
        "resource_snapshot_id": model["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.create_existing_resource_draft(
        client, resource_route=V1_MODELS_ROUTE, resource_id=model["id"], payload=payload
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)
    asserts.assert_retrieving_draft_by_resource_id_works(
        client,
        resource_route=V1_MODELS_ROUTE,
        resource_id=model["id"],
        expected=response,
    )
    asserts.assert_creating_another_existing_draft_fails(
        client, resource_route=V1_MODELS_ROUTE, resource_id=model["id"]
    )

    # test modification
    payload = {"name": new_name, "description": description}
    expected = {
        "user_id": auth_account["id"],
        "group_id": model["group"]["id"],
        "resource_id": model["id"],
        "resource_snapshot_id": model["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.modify_existing_resource_draft(
        client,
        resource_route=V1_MODELS_ROUTE,
        resource_id=model["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)

    # test deletion
    actions.delete_existing_resource_draft(
        client, resource_route=V1_MODELS_ROUTE, resource_id=model["id"]
    )
    asserts.assert_existing_draft_is_not_found(
        client, resource_route=V1_MODELS_ROUTE, resource_id=model["id"]
    )


def test_manage_new_model_drafts(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that drafts of model can be created and managed by the user

    Given an authenticated user, this test validates the following sequence of actions:

    - The user creates two model drafts
    - The user retrieves information about the drafts and gets the expected response
    - The user modifies the description of the model in the first draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the first draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    group_id = auth_account["groups"][0]["id"]
    drafts = {
        "draft1": {"name": "model1", "description": "new model"},
        "draft2": {"name": "model2", "description": None},
    }

    # test creation
    draft1_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft1"],
    }
    draft1_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_MODELS_ROUTE,
        group_id=group_id,
        payload=drafts["draft1"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft1_response, draft1_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_MODELS_ROUTE,
        draft_id=draft1_response["id"],
        expected=draft1_response,
    )
    draft2_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft2"],
    }
    draft2_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_MODELS_ROUTE,
        group_id=group_id,
        payload=drafts["draft2"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft2_response, draft2_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_MODELS_ROUTE,
        draft_id=draft2_response["id"],
        expected=draft2_response,
    )
    asserts.assert_retrieving_drafts_works(
        client,
        resource_route=V1_MODELS_ROUTE,
        expected=[draft1_response, draft2_response],
    )

    # test modification
    draft1_mod = {"name": "draft1", "description": "new description"}
    draft1_mod_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": draft1_mod,
    }
    response = actions.modify_new_resource_draft(
        client,
        resource_route=V1_MODELS_ROUTE,
        draft_id=draft1_response["id"],
        payload=draft1_mod,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        response, draft1_mod_expected
    )

    # test deletion
    actions.delete_new_resource_draft(
        client, resource_route=V1_MODELS_ROUTE, draft_id=draft1_response["id"]
    )
    asserts.assert_new_draft_is_not_found(
        client, resource_route=V1_MODELS_ROUTE, draft_id=draft1_response["id"]
    )


def test_manage_model_snapshots(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
    registered_model_versions: dict[str, Any],
) -> None:
    """Test that different snapshots of a model can be retrieved by the user.

    Given an authenticated user and registered models, this test validates the following
    sequence of actions:

    - The user modifies a model
    - The user retrieves information about the original snapshot of the model and gets
      the expected response
    - The user retrieves information about the new snapshot of the model and gets the
      expected response
    - The user retrieves a list of all snapshots of the model and gets the expected
      response
    """
    model_to_rename = registered_models["model1"]
    modified_model = modify_model(
        client,
        model_id=model_to_rename["id"],
        new_name=model_to_rename["name"] + "modified",
        new_description=model_to_rename["description"],
    ).get_json()
    modified_model["latestVersion"] = None
    modified_model["versions"] = []
    modified_model.pop("hasDraft")
    model_to_rename.pop("hasDraft")
    model_to_rename["latestSnapshot"] = False
    model_to_rename["lastModifiedOn"] = modified_model["lastModifiedOn"]
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_MODELS_ROUTE,
        resource_id=model_to_rename["id"],
        snapshot_id=model_to_rename["snapshot"],
        expected=model_to_rename,
    )
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_MODELS_ROUTE,
        resource_id=modified_model["id"],
        snapshot_id=modified_model["snapshot"],
        expected=modified_model,
    )
    expected_snapshots = [model_to_rename, modified_model]
    asserts.assert_retrieving_snapshots_works(
        client,
        resource_route=V1_MODELS_ROUTE,
        resource_id=model_to_rename["id"],
        expected=expected_snapshots,
    )


def test_tag_model(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that tags can applied to models.

    Given an authenticated user, registered models, and registered tags this test
    validates the following sequence of actions:

    - The user appends tags to a model
    - The user retrieves information about the tags of the model and gets
      the expected response
    - The user removes a tag from a model
    - The user retrieves information about the tags of the model and gets
      the expected response
    - The user modifies a tag from a model
    - The user retrieves information about the tags of the model and gets
      the expected response
    - The user removes all tags from a model
    - The user attempts to retrieve information about the tags of the model and gets
      the gets the expected response of no tags
    """
    model = registered_models["model1"]
    tags = [tag["id"] for tag in registered_tags.values()]

    # test append
    response = actions.append_tags(
        client,
        resource_route=V1_MODELS_ROUTE,
        resource_id=model["id"],
        tag_ids=[tags[0], tags[1]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1]]
    )
    response = actions.append_tags(
        client,
        resource_route=V1_MODELS_ROUTE,
        resource_id=model["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1], tags[2]]
    )

    # test remove
    actions.remove_tag(
        client, resource_route=V1_MODELS_ROUTE, resource_id=model["id"], tag_id=tags[1]
    )
    response = actions.get_tags(
        client, resource_route=V1_MODELS_ROUTE, resource_id=model["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[2]]
    )

    # test modify
    response = actions.modify_tags(
        client,
        resource_route=V1_MODELS_ROUTE,
        resource_id=model["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[1], tags[2]]
    )

    # test delete
    response = actions.remove_tags(
        client, resource_route=V1_MODELS_ROUTE, resource_id=model["id"]
    )
    response = actions.get_tags(
        client, resource_route=V1_MODELS_ROUTE, resource_id=model["id"]
    )
    asserts.assert_tags_response_contents_matches_expectations(response.get_json(), [])


# -- Tests Model Versions --------------------------------------------------------------


def test_create_model_version(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
    registered_artifacts: dict[str, Any],
) -> None:
    """Test that model version can be correctly registered and retrieved using the API.

    Given an authenticated user, registered models, and registered artifacts this test
    validates the following sequence of actions:

    - The user registers a model version.
    - The response is valid and matches the expected values given the registration
      request.
    - The user is able to retrieve information about the model version using the version
      number.
    """
    model_id = registered_models["model1"]["id"]
    artifact_id = registered_artifacts["artifact1"]["id"]
    description = "The model version."
    model_version_response = actions.register_model_version(
        client, model_id=model_id, artifact_id=artifact_id, description=description
    ).get_json()

    assert_model_version_response_contents_matches_expectations(
        response=model_version_response,
        expected_contents={
            "model_id": model_id,
            "artifact_id": artifact_id,
            "description": description,
        },
    )

    assert_retrieving_model_version_by_version_number_works(
        client,
        model_id=model_id,
        version_number=model_version_response["versionNumber"],
        expected=model_version_response,
    )


def test_model_versions_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
    registered_model_versions: dict[str, Any],
) -> None:
    """Test that all model versions can be retrieved.

    Given an authenticated user, registered model, and registered model versions
    this test validates the following sequence of actions:

    - The user is able to retrieve a list of all registered model versions.
    - The returned list of models matches the full list of registered models.
    """
    model_id = registered_models["model1"]["id"]
    model_version_expected_list = list(registered_model_versions.values())[::4]
    assert_retrieving_model_versions_works(
        client, model_id=model_id, expected=model_version_expected_list
    )


def test_model_version_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
    registered_model_versions: dict[str, Any],
) -> None:
    """Test that model versions can be queried with a search term.

    Given an authenticated user, registered model, and registered model versions
    this test validates the following sequence of actions:

    - The user is able to retrieve a list of all registered model versions with various
      queries.
    - The returned list of model versions matches the expected matches from the query.
    """
    model_id = registered_models["model1"]["id"]
    model_version_expected_list = list(registered_model_versions.values())[::4]
    assert_retrieving_model_versions_works(
        client,
        model_id=model_id,
        expected=model_version_expected_list,
        search="description:*version*",
    )
    model_id = registered_models["model2"]["id"]
    model_version_expected_list = list(registered_model_versions.values())[1:-3]
    assert_retrieving_model_versions_works(
        client,
        model_id=model_id,
        expected=model_version_expected_list,
        search="description:*version*",
    )
    model_id = registered_models["model2"]["id"]
    model_version_expected_list = [
        list(registered_model_versions.values())[n] for n in (1, 3)
    ]
    assert_retrieving_model_versions_works(
        client,
        model_id=model_id,
        expected=model_version_expected_list,
        search="*",
    )


def test_modify_model_version(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
    registered_model_versions: dict[str, Any],
) -> None:
    model_id = registered_models["model1"]["id"]
    model_version_number_to_modify = registered_model_versions["version1"][
        "versionNumber"
    ]
    new_description = "The new description"
    modify_model_version(
        client,
        model_id=model_id,
        version_number=model_version_number_to_modify,
        new_description=new_description,
    )
    assert_model_version_description_matches_expected_description(
        client,
        model_id=model_id,
        version_number=model_version_number_to_modify,
        expected_description=new_description,
    )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from typing import Dict, List

import entrypoints
import structlog
from entrypoints import EntryPoint
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()


_GENERICS: List[str] = ["estimator_predict", "fit_estimator"]
_GENERICS_ENTRYPOINTS: Dict[str, Dict[str, EntryPoint]] = {}


def _register(generic: str, entrypoint: EntryPoint) -> None:
    generic_registry = _GENERICS_ENTRYPOINTS.setdefault(generic, {})

    try:
        generic_registry[entrypoint.name] = entrypoint.load()

    except (AttributeError, ImportError):
        LOGGER.exception(
            "Failed to register dispatch method for generic", name=f"{generic}()"
        )


def register_entrypoints() -> None:
    for generic in _GENERICS:
        for entrypoint in entrypoints.get_group_all(f"dioptra.generics.{generic}"):
            _register(generic=generic, entrypoint=entrypoint)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import importlib
from functools import wraps
from typing import Any, Callable, Optional, Type, TypeVar, cast

import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.exceptions.base import BaseOptionalDependencyError

LOGGER: BoundLogger = structlog.stdlib.get_logger()

Exc = Type[BaseOptionalDependencyError]
Function = Callable[..., Any]
T = TypeVar("T", bound=Function)


def require_package(
    name: str,
    exc_message: Optional[str] = None,
    exc_type: Exc = BaseOptionalDependencyError,
) -> Callable[[T], T]:
    def decorator(func: T) -> T:
        error_msg: str = f'The package "{name}" is required to use {func.__name__}'
        exc: BaseOptionalDependencyError = (
            exc_type(exc_message) if exc_message else exc_type(error_msg)
        )

        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                importlib.import_module(name=name)

            except ModuleNotFoundError:
                LOGGER.exception(error_msg, args=args, kwargs=kwargs)
                raise exc

            return func(*args, **kwargs)

        return cast(T, wrapper)

    return decorator
